#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <map>
#include <set>
#include <string>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

struct Item
{
    Item(const std::string& key) 
        : m_key(key), m_sellPrice(0), m_buyPrice(0), 
          m_makePrice(0.0), m_runtHours(0.0) { } 

    double getPrice();
    double getHours();

    std::string m_key;
    std::string m_name;

    double m_sellPrice;
    double m_buyPrice;
    
    double m_makePrice;
    double m_runtHours;
};

double Item::getPrice()
{
    if(m_makePrice == 0.0) return m_buyPrice;
    if(m_buyPrice == 0.0) return m_makePrice;
    return (m_buyPrice < m_makePrice)?m_buyPrice:m_makePrice;
}

double Item::getHours()
{
    if(m_makePrice == 0.0) return 0.0; // Item has no viable make option.
    if(m_buyPrice == 0.0 || (m_makePrice < m_buyPrice)) return m_runtHours;
    return 0.0;
}

struct ItemTag
{
    ItemTag(Item* item, int qty) : m_item(item), m_qty(qty) { }
    Item* m_item;
    int   m_qty;
};

typedef std::list<ItemTag> ItemTagList;
typedef std::map<std::string, Item*> ItemMap;

struct Manufacture
{
    Manufacture(const std::string& key)
        : m_key(key), m_makeTime(0), m_makeCost(0),
          m_totalCost(0.0), m_totalTime(0.0), m_sellPrice(0.0) { }

    void addBase(const std::string& reqs);
    bool analyze();
    void print();
    double mppr();

    std::string m_key;
    std::string m_name;
    
    int m_makeTime;
    int m_makeCost;
    
    double m_totalCost;
    double m_totalTime;

    double m_sellPrice;

    std::set<std::string> m_facilities;
    std::set<std::string> m_techs;
    
    ItemTagList m_materials;
    ItemTagList m_results;
};

typedef std::map<std::string, Manufacture*> MfgrMap;

struct MfgSet
{
    Item* addItem(const std::string& key);
    void verify_results(Manufacture* mfgr);
    int parseManufactures(const std::string& fileName);
    void sizes() const;
    int parseItems(const std::string& fileName);
    int translateItems(const std::string& fileName);

    MfgrMap m_mfgrs;
    ItemMap m_items;
};

void Manufacture::addBase(const std::string& reqs)
{
    size_t pos = 0;
    while(pos != std::string::npos)
    {
        pos = reqs.find_first_not_of("[], ", pos);
        if(pos == std::string::npos) return;
        size_t endpos = reqs.find_first_of("[], ", pos);
        if(endpos == std::string::npos) return;
        m_facilities.insert(reqs.substr(pos, endpos - pos));
        pos = endpos;
    }
}

bool hasValue(const std::string& line, const char* tag, std::string& val)
{
    size_t pos = line.find(tag);
    if(pos == std::string::npos) return false;
    pos += strlen(tag);
    pos = line.find_first_not_of(' ', pos);
    if(pos != std::string::npos) val = line.substr(pos);
    return true;
}

int itemQty(const std::string& line, std::string& val)
{
    size_t pos = line.find("STR_");
    if(pos == std::string::npos) return 0;
    size_t colpos = line.find(':', pos);
    if(colpos == std::string::npos) return 0;
    val = line.substr(pos, colpos - pos);
    return atoi(line.substr(colpos + 2).c_str());
}

bool Manufacture::analyze()
{
    bool changed = false;
    
    double totalCost = (5000.0 * m_makeTime) / 720.0 + m_makeCost;
    double totalTime = m_makeTime;

    ItemTagList::iterator mt;
    for(mt = m_materials.begin(); mt != m_materials.end(); ++mt)
    {
        // price rules:
        // if item has buy price or true make price, get lowest extant
        Item* item = mt->m_item;
        if(totalCost > 0.0)
        {
            double itemPrice = item->getPrice();
            if(itemPrice > 0.0)  
            {
                totalCost += (itemPrice * mt->m_qty);
                totalTime += (item->getHours() * mt->m_qty);
            }
            else totalCost = 0.0;
        }
    }
    if(totalCost > 0.0 && (m_totalCost == 0.0 || totalCost < m_totalCost))
    {
        m_totalCost = totalCost;
        m_totalTime = totalTime;
        changed = true;
    }

    bool item_change = false;

    if(changed || m_sellPrice == 0.0)
    {
        m_sellPrice = 0.0;
        for(mt = m_results.begin(); mt != m_results.end(); ++mt)
        {
            Item* item = mt->m_item;
            m_sellPrice += (item->m_sellPrice * mt->m_qty);
        }
        // Distribute cost to make if available, proportionate to cost to sell
        if(changed)
        {
            double makeRatio = m_totalCost / m_sellPrice;
            
            for(mt = m_results.begin(); mt != m_results.end(); ++mt)
            {
                Item* item = mt->m_item;
                if(m_totalCost > 0.0)
                {
                    double newMakePrice = item->m_sellPrice * makeRatio;
                    if(item->m_makePrice == 0.0 || item->m_makePrice > newMakePrice) 
                    {
                        item->m_makePrice = newMakePrice;
                        item->m_runtHours = m_totalTime / mt->m_qty;
                        item_change = true;
                    }
                }
            }
        }
    }
    return item_change;
}

void Manufacture::print()
{
    std::cout << m_key << ',' << m_name << ',' << m_totalCost << ',' << m_sellPrice << std::endl;
}

// Calculate monthly profit with 50 runts, including facility costs.
double Manufacture::mppr()
{
    if(m_totalCost == 0.0) return 0.0;
    double runt50 = (720.00 / m_totalTime) * (m_sellPrice - m_totalCost) * 50;
    double fcost = 15000.0;  //2x living quarters
    int wspace = 0;
    std::set<std::string>::iterator ft = m_facilities.begin();
    while(ft != m_facilities.end())
    {
        if(*ft == "ARMS") 
        {
            fcost += 200000,0;
            wspace += 10;
        }
        else if(*ft == "SPA") 
        {
            fcost += 100000,0;
        }
        else if(*ft == "CULT") 
        {
            fcost += 10000,0;
        }
        else if(*ft == "ALKO") 
        {
            fcost += 10000,0;
            wspace += 10;
        }     
        else if(*ft == "MINT") 
        {
            fcost += 40000,0;
            wspace += 10;
        }    
        else if(*ft == "REFI") 
        {
            fcost += 25000,0;
            wspace += 10;
        }  
        else if(*ft == "SUMM") 
        {
            fcost += 30000,0;
            wspace += 10;
        }  
        else if(*ft == "WELL" || *ft == "SHOP") 
        {
            fcost += 15000,0;
            wspace += 20;
        }  
        else if(*ft == "WORKS") 
        {
            fcost += 75000,0;
            wspace += 30;
            if(m_facilities.find("SHOP") != m_facilities.end()) 
            {
                fcost -= 15000,0;
                wspace -= 20;
            }
        }  
        else if(*ft == "PRINT") 
        {
            fcost += 125000,0;
            wspace += 100;
            if(m_facilities.find("WORKS") != m_facilities.end())
            {
                fcost -= 75000,0;
                wspace -= 30;
            }
        }
        else if(*ft == "SUMM") 
        {
            fcost += 30000,0;
            wspace += 10;
        }   
        else if(*ft == "MED" || *ft == "SUR") 
        {
            fcost += 70000,0;                
        }  
        else if(*ft == "LIB") 
        {
            fcost += 45000,0;
        } 
        else if(*ft == "CPU2" || *ft == "CPU") 
        {
            fcost += 150000,0;
        } 
        ++ft;
        while(wspace < 50)
        {
            fcost += 15000,0; // Add more extractors
            wspace += 20;
        }
    }
    return (runt50 - fcost);
}

Item* MfgSet::addItem(const std::string& key)
{
    ItemMap::iterator it = m_items.find(key);
    if(it != m_items.end()) return it->second;
    Item* item = new Item(key);
    m_items[key] = item;
    return item;
}

void MfgSet::verify_results(Manufacture* mfgr)
{
    // If manufacture had no produced item specified
    // Its result is a single quantity of itself
    if(mfgr->m_results.empty())
    {
        Item* item = addItem(mfgr->m_key); 
        mfgr->m_results.push_back(ItemTag(item, 1));
    }
}

void trimTrailing(std::string& line)
{
    size_t pos = line.find_last_not_of("\n\r\t ");
    if(pos != std::string::npos && pos < line.size())
    {
        line.erase(pos+1);
    }
}

int MfgSet::parseManufactures(const std::string& fileName)
{
    int parsed = 0;
    std::ifstream ruleFile(fileName.c_str());

    if (!ruleFile)
    {
        std::cerr << "Unable to open " << fileName << std::endl;
        return parsed;
    }

    std::string sLine;
    while ( getline(ruleFile, sLine))
    {
        if(sLine.find("manufacture:")  == 0) break;
    }

    Manufacture* mfgr = NULL;
    bool techs = false;
    bool required = false;
    bool produced = false;
    
    while ( getline(ruleFile, sLine))
    {
        trimTrailing(sLine);
        if(sLine[0] != ' ' && sLine[0] != '#') break; // Next section start
        std::string val;
        if(hasValue(sLine, "- name:", val)) // start new manufacture
        { 
            if(mfgr) verify_results(mfgr);

            ++parsed;           
            mfgr = new Manufacture(val);
            m_mfgrs[val] = mfgr;
            techs = false;
            required = false;
            produced = false;
            continue;
        }
        if(mfgr)
        {
            if(techs)
            {
                if(hasValue(sLine, "-", val))
                {
                    mfgr->m_techs.insert(val);
                    continue;
                }
                else techs = false;
            }
            else if(required)
            {
                int qty = itemQty(sLine, val);
                if(qty > 0)
                {                    
                    Item* item = addItem(val);
                    mfgr->m_materials.push_back(ItemTag(item, qty));
                    continue;
                }
                else required = false;
            }
            else if(produced)
            {
                int qty = itemQty(sLine, val);
                if(qty > 0)
                {
                    Item* item = addItem(val);
                    mfgr->m_results.push_back(ItemTag(item, qty));
                    continue;
                }
                else produced = false;
            }
            // Fell through here
            if(hasValue(sLine, "requires:", val))
            {
                techs = true;                
            }
            else if(hasValue(sLine,"requiredItems:", val))
            {
                required = true;
            }
            else if(hasValue(sLine,"producedItems:", val))
            {
                produced = true;
            }
            else if(hasValue(sLine,"time:", val))
            {
                mfgr->m_makeTime = atoi(val.c_str());
            }
            else if(hasValue(sLine,"cost:", val))
            {
                mfgr->m_makeCost = atoi(val.c_str());
            }
            else if(hasValue(sLine,"requiresBaseFunc:", val))
            {
                mfgr->addBase(val);
            }
        }
    }
    
    if(mfgr) verify_results(mfgr);

    return parsed;
}
void MfgSet::sizes() const
{
    std::cout << "Manufactures: " << m_mfgrs.size() << ", Items: " << m_items.size() << std::endl;
}

int MfgSet::parseItems(const std::string& fileName)
{
    int parsed = 0;
    int total = 0;
    std::ifstream ruleFile(fileName.c_str());

    if (!ruleFile)
    {
        std::cerr << "Unable to open " << fileName << std::endl;
        return parsed;
    }

    std::string sLine;
    while ( getline(ruleFile, sLine))
    {
        if(sLine.find("items:") == 0) break;
    }

    Item* item = NULL;
    while ( getline(ruleFile, sLine))
    {
        trimTrailing(sLine);
        if(sLine[0] != ' ' && sLine[0] != '#' && sLine[0] != 'c') 
        {
            break; // Next section start
        }
        std::string val;
        if(hasValue(sLine, "- type:", val)) // start new item
        { 
            ++total;
            ItemMap::iterator it = m_items.find(val);
            if(it != m_items.end()) 
            {
                item = it->second;
                ++parsed;       
            }
            else 
            {
                item = NULL;
            }
            continue;
        }
        if(item)
        {
            if(hasValue(sLine,"costSell:", val))
            {
                item->m_sellPrice = atoi(val.c_str());
            }
            else if(hasValue(sLine,"costBuy:", val))
            {
                item->m_buyPrice = atoi(val.c_str());
            }
        }
    }

    std::cout << "Total items read: " << total << std::endl;

    return parsed;
}

int MfgSet::translateItems(const std::string& fileName)
{
    int parsed = 0;
    int itrans = 0;
    int mtrans = 0;
    std::ifstream langFile(fileName.c_str());
    if (!langFile)
    {
        std::cerr << "Unable to open " << fileName << std::endl;
        return parsed;
    }

    std::string sLine;
    while ( getline(langFile, sLine))
    {
        trimTrailing(sLine);
        size_t pos = sLine.find("STR_");
        if(pos == std::string::npos) continue;
        size_t colpos = sLine.find(':', pos);
        if(colpos == std::string::npos) continue;
        size_t q1pos = sLine.find_first_not_of(": ", colpos);
        if(q1pos == std::string::npos) continue;
        size_t q2pos;
        if(sLine[q1pos] == '"') // Value is in quotes
        {
            ++q1pos;
            q2pos = sLine.find_first_of('"', q1pos);
            if(q2pos == std::string::npos) continue;
        }
        else // No quotes
        {
            q2pos = sLine.find_last_not_of(" \n\r\t");
            if(q2pos < q1pos) continue;
            q2pos++;
        }
        ++parsed;
        std::string key = sLine.substr(pos, colpos - pos);
        std::string val = sLine.substr(q1pos, q2pos - q1pos);

        ItemMap::iterator it = m_items.find(key);
        if(it != m_items.end())
        {
            ++itrans;
            it->second->m_name = val;
            //std::cout << "Item [" << key << "] name [" << val << "]" << std::endl;
        }
        
        MfgrMap::iterator mt =  m_mfgrs.find(key);
        if(mt != m_mfgrs.end())
        {
            ++mtrans;
            mt->second->m_name = val;
        }
    }

    ItemMap::iterator it = m_items.begin();
    while(it != m_items.end())
    {
        if(it->second->m_name.empty()) 
        {
            ++itrans;
            it->second->m_name = it->first.substr(5);
        }
        ++it;
    }

    MfgrMap::iterator mt =  m_mfgrs.begin();
    while(mt != m_mfgrs.end())
    {
        if(mt->second->m_name.empty()) 
        {
            ++mtrans;
            mt->second->m_name = mt->first.substr(5);
        }
        ++mt; 
    }   
    
    std::cout << "Mfgrs translated: " << mtrans << ", Items translated: " << itrans << std::endl;

    return parsed;    
}



int main(int argc, char **argv)
    
{
    // Steps:
    // Open the rules file
    // Scan and parse all manufactures, building item map;
    // Scan and parse items, any item in item map, assign buy and sell prices.
    // Scan lang file, apply translations to items in map.
    // Analyze and calculate prices.
    MfgSet mfgSet;

    std::string ruleFileName("Piratez.rul");
    if(argc > 1)
    {
        ruleFileName.assign(argv[1]);
    }

    int nparsed = mfgSet.parseManufactures(ruleFileName);
    mfgSet.sizes();
    nparsed = mfgSet.parseItems(ruleFileName);
    std::cout << "Got info on "  << nparsed << " items" << std::endl;
    std::string langFileName("Piratez_lang.rul");
    nparsed = mfgSet.translateItems(langFileName);
    std::cout << "Translated "  << nparsed << " lines" << std::endl;

    // Analisys iterations
    for(int i = 0; i < 50; ++i)
    {
        int nchanged = 0;
        MfgrMap::iterator mt = mfgSet.m_mfgrs.begin();
        while(mt != mfgSet.m_mfgrs.end())
        {
            if(mt->second->analyze()) ++nchanged;
            ++mt;
        }    
        if(nchanged == 0) break;
    }

    typedef std::map<double, Manufacture*> ProfitMap;

    ProfitMap profitMap;
    MfgrMap::iterator mt = mfgSet.m_mfgrs.begin();
    while(mt != mfgSet.m_mfgrs.end())
    {
        double mppr = mt->second->mppr();
        if(mppr > 0.0) profitMap[mppr] = mt->second;
        ++mt;
    }

    ProfitMap::reverse_iterator pt = profitMap.rbegin();
    while(pt != profitMap.rend())
    {
        std::cout << pt->first << " : ";
        pt->second->print();
        ++pt;
    }

    return 0;
}
